/*
 * @(#RadioBox.java
 *
 * Copyright (c) 1997 Symantec Corporation. All Rights Reserved.
 *
 */


package symantec.itools.db.awt;

import java.util.*;
import java.sql.*;
import java.lang.*;
import symantec.itools.db.net.*;
import symantec.itools.db.pro.*;
import symantec.itools.db.beans.binding.Name;

/**
 * A dbAWARE CheckboxGroup component.
 * <p>
 * This component is similar to the basic CheckboxGroup component, but can
 * be "bound" to a projection within a relation view
 * so that it automatically gets and sets the values in that relation.
 * <p>
 *
 * @see java.awt.CheckboxGroup
 */
public class RadioBox extends java.awt.CheckboxGroup implements ProjectionBean
{
    private ProjectionBeanHelper m_Helper;

    /**
     * A constant value indicating how an empty string will be saved in the database.
     */
    public static final String Default = ProjectionBeanHelper.DEFAULT;
    /**
     * A constant value indicating how an empty string will be saved in the database.
     */
    public static final String Blank = ProjectionBeanHelper.BLANK;
    /**
     * A constant value indicating how an empty string will be saved in the database.
     */
    public static final String Null = ProjectionBeanHelper.NULL;

    private Vector rbkVector = new Vector();

    /**
     * Constructs a new RadioBox.
     * @param the container with the Checkboxes to be managed by this RadioBox
     */
    public RadioBox()
    {
        super();
        m_Helper = new ProjectionBeanHelper(this);
    }

    /**
     * Binds this component to a given projection within the specified
     * relation view.
     *
     * @param relView the relation view to bind with
     * @param projection the projection in relView to bind with
     */
    public void setBinding(RelationView relView, String projection)
    {
        m_Helper.setBinding(relView, projection);
    }

    /**
     * Returns the projection in the RelationView that this component is bound with.
     * @see #setBinding
     * @see #setProjection
     */
    public String getProjection() {
        return m_Helper.getProjection();
    }

    /**
     * Binds this component to the given projection within the RelationView
     * the component is currently bound with.
     * @see #setBinding
     * @see #getProjection
     * @see #getRelationView
     */
    public void setProjection(String projection) {
        m_Helper.setProjection(projection);
    }

    /**
     * Gets the RelationView that this component is bound with.
     * @return the RelationView currently bound with
     * @see #setRelationView
     * @see #setBinding
     * @see #getProjection
     */
    public RelationView getRelationView() {
        return m_Helper.getRelationView();
    }

    /**
     * Sets the RelationView that this component is bound with.
     * @param value the RelationView to bind with
     * @see #getRelationView
     * @see #setBinding
     * @see #setProjection
     */
    public void setRelationView(RelationView value) {
        m_Helper.setRelationView(value);
    }

    /**
     * Specifies how an empty string will be set when updating data on
     * the dbANYWHERE server.
     *
     * @param blank one of "DEFAULT", "NULL", or "BLANK"
     *
     * @see symantec.itools.db.pro.ProjBinder#setValueFromString(java.lang.String, int, int)
     */
    public void setTreatBlankAs(String value) {
        m_Helper.setTreatBlankAsString(value);
    }

    /**
     * Indicates when the component commits its changes.
     * @see #setDynamicUpdate
     */
    public boolean getDynamicUpdate() {
        return m_Helper.getDynamicUpdate();
    }

    /**
     * Sets when the component commits its changes.
     * @param value the new dynamic update mode value
     * @see #getDynamicUpdate
     */
    public void setDynamicUpdate(boolean value) {
        m_Helper.setDynamicUpdate(value);
    }


    /**
     * Sets whether the data value of this component may be modified.
     * @param value <code>true</code> if the value may not be modified,
     * <code>false</code>if the value may be modified
     */
    public void setReadOnly(boolean value) {
    }

    /**
     * Sets the value of this component to the given value.
     * @param value the new component value
     * @see #getData
     */
    public void setData(Object value) {
        String text="";
        if(value!=null) text = value.toString();
        Enumeration itemElements = rbkVector.elements();
        try
        {
            while (itemElements.hasMoreElements())
            {
                java.awt.Checkbox cb = (java.awt.Checkbox)itemElements.nextElement();
                if (cb != null && cb.getLabel().equals(text)) {
                    cb.setState(true);
                    return;
                }
            }
	        setCurrent(null);
        }
        catch (Exception Ex) {
            m_Helper.raiseException(Ex);
        }
    }

    /**
     * Gets the value of this component.
     * @return the current component value
     * @see #setData
     */
    public Object getData() {
        String text = null;
        Enumeration itemElements = rbkVector.elements();

        try
        {
            while (itemElements.hasMoreElements())
            {
                java.awt.Checkbox cb = (java.awt.Checkbox)itemElements.nextElement();
                if (cb != null && cb.getState()) {
                    text = cb.getLabel();
                    break;
                }
            }
        }
        catch (Exception Ex) {
            m_Helper.raiseException(Ex);
        }
        return text;
    }

    /**
     * Gets whether this component saves its value as a text String.
     * @return <code>true</code> if the value is saved as text,
     * <code>false</code> otherwise
     */
    public boolean isTextBased() {
        return true;
    }

    /**
     * Gets the number of digits to the right of the decimal point for
     * this component's value.
     * @return the number of digits to the right of the decimal point
     */
    public int getScale() {
        return ProjBinder.DEFAULTSCALE;
    }

    /**
     * Registers the standard event listener(s) for this component.
     */
    public void registerListeners() {
    }

    void addCheckbox(java.awt.Checkbox checkbox) {
        rbkVector.addElement(checkbox);
        m_Helper.notifyDataChange();
        checkbox.addItemListener(m_Helper);
        checkbox.addFocusListener(m_Helper);
    }
    /**
     * Sets the name of the data item to bind this component to.
     * @param name the data item name, like "MyTable@MyColumn"
     * @see #getDataBinding
     */
    public void setDataBinding(String name)
    {
        m_Helper.setDataBinding(new Name(name));
    }

    /**
     * Gets the name of the data item this component is bound to.
     * @returns the data item name, like "MyTable@MyColumn"
     * @see #setDataBinding
     */
    public String getDataBinding()
    {
        return  m_Helper.getDataBinding().getFullName();
    }

     /**
     * Sets the value of the Empty Means Null property.
     * This determines how an empty string will be saved when updating data on
     * the database server.
     * A value of <code>true</code>indicates an empty string will be saved as a
     * SQL NULL value. Otherwise it will be saved as a blank value.
     * @param propertyValue <code>true</code> for NULL, <code>false</code> for BLANK
     * @see #getEmptyMeansNull
     */
    public void setEmptyMeansNull(boolean propertyValue)
    {
        m_Helper.setEmptyMeansNull(propertyValue);
    }

     /**
     * Gets the value of the Empty Means Null property.
     * This determines how an empty string will be saved when updating data on
     * the database server.
     * A value of <code>true</code>indicates an empty string will be saved as a
     * SQL NULL value. Otherwise it will be saved as a blank value.
     * @return <code>true</code> for NULL, <code>false</code> for BLANK
     * @see #setEmptyMeansNull
     */
    public boolean getEmptyMeansNull()
    {
        return m_Helper.getEmptyMeansNull();
    }

    /**
     * For backward compatbilty only
     * @param the container with the Checkboxes to be managed by this RadioBox
     */
    public RadioBox(java.awt.Container c)
    {
        super();
        m_Helper = new ProjectionBeanHelper(this);
    }
}
